@@ -9,11 +9,13 @@
 #define __arconv_h_
 
 #include <glib.h>
+#include "pango.h"
+#include "pangox.h"
 
 /*
  * arabic_reshape: reshapes string ( ordered left-to right visual order )
  *   len :  before: is the length of the string 
- *          after : number of  nun-NULL characters
+ *          after : number of  non-NULL characters
  *
  */
 typedef enum 
@@ -25,7 +27,18 @@ typedef enum
     ar_unifont  = 0x40,    ar_naqshfont        = 0x80
 } arabic_level;
 
-void arabic_reshape(int* len,gunichar* string,arabic_level level);
+typedef struct
+{
+    PangoXSubfont subfonts[3];
+    arabic_level  level;
+} ArabicFontInfo;
+
+/*  len   : beforehand: #chars in string
+ *               after: #chars in text
+ *  string: original-string
+ *  text  : output-text
+ */
+void arabic_reshape(int* len,const char* text,gunichar* string,arabic_level level);
 int  arabic_isvowel(gunichar s);
 
 #endif
