@@ -8,13 +8,14 @@
 #define __lboxfont_h_
 #include "pango.h"
 #include "pangox.h"
+#include "arconv.h"
 
 /* 
- * lboxfont must point to valid memory for this to work.
+ *  create an arabic_fontstruct for the langbox-module
+ *  returns: NULL on failure
  */
-int
-arabic_lboxinit(PangoFont  *font,PangoXSubfont* lboxfonts);
-/* a return value of 0 means this has failed */
+ArabicFontInfo*
+arabic_lboxinit(PangoFont  *font);
 
 
 /* glyph2 is the next glyph in line; if there is none, put in NULL 
