@@ -2,19 +2,20 @@
  * 
  * (C) 2000 K. Koehler  <koehler@or.uni-bonn.de>
  *
- * This file provides a mapping unicode <- mulefont
+ * This file provides a mapping unicode <- urdu nq-font
  */
 #ifndef __nqfont_h_
 #define __nqfont_h_
 #include "pango.h"
 #include "pangox.h"
+#include "arconv.h"
 
 /* 
- * nqfont must point to valid memory for this to work.
+ *  create an arabic_fontstruct for the urdu_naqshfont-module
+ *  returns: NULL on failure
  */
-int
-urdu_naqshinit(PangoFont  *font,PangoXSubfont* nqfont);
-/* a return value of 0 means this has failed */
+ArabicFontInfo*
+urdu_naqshinit(PangoFont  *font);
 
 
 /* glyph2 is the next glyph in line; if there is none, put in NULL 
@@ -23,3 +24,5 @@ void
 urdu_naqsh_recode(PangoXSubfont* subfont,int* glyph,int* glyph2,
                    PangoXSubfont* nqfont);
 #endif
+
+
