@@ -604,16 +604,8 @@ hangul_engine_shape (PangoFont        *font,
     if (!find_subfont (font, secondary_charset, G_N_ELEMENTS (secondary_charset), &subfont, &render_func))
       if (!find_subfont (font, fallback_charsets, G_N_ELEMENTS (fallback_charsets), &subfont, &render_func))
 	{
-	  PangoGlyph unknown_glyph = pango_x_get_unknown_glyph (font);
-	  
 	  n_chars = g_utf8_strlen (text, length);
-	  pango_glyph_string_set_size (glyphs, n_chars);
-
-	  for (i=0; i<n_chars; i++)
-	    set_glyph (glyphs, i, font,
-		       PANGO_X_GLYPH_SUBFONT (unknown_glyph), 
-		       PANGO_X_GLYPH_INDEX (unknown_glyph));
-
+	  pango_x_fallback_shape (font, glyphs, text, n_chars);
 	  return;
 	}
 
