@@ -19,6 +19,8 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <string.h>
+
 #include <pango/pango-attributes.h>
 
 struct _PangoAttrList
@@ -37,13 +39,16 @@ struct _PangoAttrIterator
 };
 
 static PangoAttribute *pango_attr_color_new  (const PangoAttrClass *klass,
-					      guint16         red,
-					      guint16         green,
-					      guint16         blue);
+                                              guint16               red,
+                                              guint16               green,
+                                              guint16               blue);
 static PangoAttribute *pango_attr_string_new (const PangoAttrClass *klass,
-					      const char     *str);
+                                              const char           *str);
 static PangoAttribute *pango_attr_int_new    (const PangoAttrClass *klass,
-					      int             value);
+                                              int                   value);
+static PangoAttribute *pango_attr_float_new  (const PangoAttrClass *klass,
+                                              double                value);
+
 
 /**
  * pango_attr_type_register:
@@ -72,9 +77,15 @@ pango_attr_type_register (const gchar *name)
 PangoAttribute *
 pango_attribute_copy (const PangoAttribute *attr)
 {
+  PangoAttribute *result;
+
   g_return_val_if_fail (attr != NULL, NULL);
 
-  return attr->klass->copy (attr);
+  result = attr->klass->copy (attr);
+  result->start_index = attr->start_index;
+  result->end_index = attr->end_index;
+
+  return result;
 }
 
 /**
@@ -92,7 +103,7 @@ pango_attribute_destroy (PangoAttribute *attr)
 }
 
 /**
- * pango_attribute_compare:
+ * pango_attribute_equal:
  * @attr1: a #PangoAttribute
  * @attr2: another #PangoAttribute
  * 
@@ -103,8 +114,8 @@ pango_attribute_destroy (PangoAttribute *attr)
  * Return value: %TRUE if the two attributes have the same value.
  **/
 gboolean
-pango_attribute_compare (const PangoAttribute *attr1,
-			 const PangoAttribute *attr2)
+pango_attribute_equal (const PangoAttribute *attr1,
+		       const PangoAttribute *attr2)
 {
   g_return_val_if_fail (attr1 != NULL, FALSE);
   g_return_val_if_fail (attr2 != NULL, FALSE);
@@ -112,7 +123,7 @@ pango_attribute_compare (const PangoAttribute *attr1,
   if (attr1->klass->type != attr2->klass->type)
     return FALSE;
 
-  return attr1->klass->compare (attr1, attr2);
+  return attr1->klass->equal (attr1, attr2);
 }
 
 static PangoAttribute *
@@ -129,8 +140,8 @@ pango_attr_string_destroy (PangoAttribute *attr)
 }
 
 static gboolean
-pango_attr_string_compare (const PangoAttribute *attr1,
-			   const PangoAttribute *attr2)
+pango_attr_string_equal (const PangoAttribute *attr1,
+			 const PangoAttribute *attr2)
 {
   return strcmp (((PangoAttrString *)attr1)->value, ((PangoAttrString *)attr2)->value) == 0;
 }
@@ -162,7 +173,7 @@ pango_attr_lang_new (const char *lang)
     PANGO_ATTR_LANG,
     pango_attr_string_copy,
     pango_attr_string_destroy,
-    pango_attr_string_compare
+    pango_attr_string_equal
   };
 
   g_return_val_if_fail (lang != NULL, NULL);
@@ -185,7 +196,7 @@ pango_attr_family_new (const char *family)
     PANGO_ATTR_FAMILY,
     pango_attr_string_copy,
     pango_attr_string_destroy,
-    pango_attr_string_compare
+    pango_attr_string_equal
   };
 
   g_return_val_if_fail (family != NULL, NULL);
@@ -209,8 +220,8 @@ pango_attr_color_destroy (PangoAttribute *attr)
 }
 
 static gboolean
-pango_attr_color_compare (const PangoAttribute *attr1,
-			  const PangoAttribute *attr2)
+pango_attr_color_equal (const PangoAttribute *attr1,
+			const PangoAttribute *attr2)
 {
   const PangoAttrColor *color_attr1 = (const PangoAttrColor *)attr1;
   const PangoAttrColor *color_attr2 = (const PangoAttrColor *)attr2;
@@ -254,7 +265,7 @@ pango_attr_foreground_new (guint16 red,
     PANGO_ATTR_FOREGROUND,
     pango_attr_color_copy,
     pango_attr_color_destroy,
-    pango_attr_color_compare
+    pango_attr_color_equal
   };
 
   return pango_attr_color_new (&klass, red, green, blue);
@@ -279,7 +290,7 @@ pango_attr_background_new (guint16 red,
     PANGO_ATTR_BACKGROUND,
     pango_attr_color_copy,
     pango_attr_color_destroy,
-    pango_attr_color_compare
+    pango_attr_color_equal
   };
 
   return pango_attr_color_new (&klass, red, green, blue);
@@ -300,8 +311,8 @@ pango_attr_int_destroy (PangoAttribute *attr)
 }
 
 static gboolean
-pango_attr_int_compare (const PangoAttribute *attr1,
-			  const PangoAttribute *attr2)
+pango_attr_int_equal (const PangoAttribute *attr1,
+		      const PangoAttribute *attr2)
 {
   const PangoAttrInt *int_attr1 = (const PangoAttrInt *)attr1;
   const PangoAttrInt *int_attr2 = (const PangoAttrInt *)attr2;
@@ -320,6 +331,41 @@ pango_attr_int_new (const PangoAttrClass *klass,
   return (PangoAttribute *)result;
 }
 
+static PangoAttribute *
+pango_attr_float_copy (const PangoAttribute *attr)
+{
+  const PangoAttrFloat *float_attr = (PangoAttrFloat *)attr;
+  
+  return pango_attr_float_new (attr->klass, float_attr->value);
+}
+
+static void
+pango_attr_float_destroy (PangoAttribute *attr)
+{
+  g_free (attr);
+}
+
+static gboolean
+pango_attr_float_equal (const PangoAttribute *attr1,
+			const PangoAttribute *attr2)
+{
+  const PangoAttrFloat *float_attr1 = (const PangoAttrFloat *)attr1;
+  const PangoAttrFloat *float_attr2 = (const PangoAttrFloat *)attr2;
+  
+  return (float_attr1->value == float_attr2->value);
+}
+
+static PangoAttribute*
+pango_attr_float_new  (const PangoAttrClass *klass,
+                       double                value)
+{
+  PangoAttrFloat *result = g_new (PangoAttrFloat, 1);
+  result->attr.klass = klass;
+  result->value = value;
+
+  return (PangoAttribute *)result;
+}
+
 /**
  * pango_attr_size_new:
  * @size: the font size, in 1000ths of a point.
@@ -335,7 +381,7 @@ pango_attr_size_new (int size)
     PANGO_ATTR_SIZE,
     pango_attr_int_copy,
     pango_attr_int_destroy,
-    pango_attr_int_compare
+    pango_attr_int_equal
   };
 
   return pango_attr_int_new (&klass, size);
@@ -356,7 +402,7 @@ pango_attr_style_new (PangoStyle style)
     PANGO_ATTR_STYLE,
     pango_attr_int_copy,
     pango_attr_int_destroy,
-    pango_attr_int_compare
+    pango_attr_int_equal
   };
 
   return pango_attr_int_new (&klass, (int)style);
@@ -377,7 +423,7 @@ pango_attr_weight_new (PangoWeight weight)
     PANGO_ATTR_WEIGHT,
     pango_attr_int_copy,
     pango_attr_int_destroy,
-    pango_attr_int_compare
+    pango_attr_int_equal
   };
 
   return pango_attr_int_new (&klass, (int)weight);
@@ -398,7 +444,7 @@ pango_attr_variant_new (PangoVariant variant)
     PANGO_ATTR_VARIANT,
     pango_attr_int_copy,
     pango_attr_int_destroy,
-    pango_attr_int_compare
+    pango_attr_int_equal
   };
 
   return pango_attr_int_new (&klass, (int)variant);
@@ -419,7 +465,7 @@ pango_attr_stretch_new (PangoStretch  stretch)
     PANGO_ATTR_STRETCH,
     pango_attr_int_copy,
     pango_attr_int_destroy,
-    pango_attr_int_compare
+    pango_attr_int_equal
   };
 
   return pango_attr_int_new (&klass, (int)stretch);
@@ -443,13 +489,13 @@ pango_attr_font_desc_destroy (PangoAttribute *attr)
 }
 
 static gboolean
-pango_attr_font_desc_compare (const PangoAttribute *attr1,
-			      const PangoAttribute *attr2)
+pango_attr_font_desc_equal (const PangoAttribute *attr1,
+			    const PangoAttribute *attr2)
 {
   const PangoAttrFontDesc *desc_attr1 = (const PangoAttrFontDesc *)attr1;
   const PangoAttrFontDesc *desc_attr2 = (const PangoAttrFontDesc *)attr2;
   
-  return pango_font_description_compare (&desc_attr1->desc, &desc_attr2->desc);
+  return pango_font_description_equal (&desc_attr1->desc, &desc_attr2->desc);
 }
 
 /**
@@ -469,7 +515,7 @@ pango_attr_font_desc_new (const PangoFontDescription *desc)
     PANGO_ATTR_FONT_DESC,
     pango_attr_font_desc_copy,
     pango_attr_font_desc_destroy,
-    pango_attr_font_desc_compare
+    pango_attr_font_desc_equal
   };
 
   PangoAttrFontDesc *result = g_new (PangoAttrFontDesc, 1);
@@ -496,7 +542,7 @@ pango_attr_underline_new (PangoUnderline underline)
     PANGO_ATTR_UNDERLINE,
     pango_attr_int_copy,
     pango_attr_int_destroy,
-    pango_attr_int_compare
+    pango_attr_int_equal
   };
 
   return pango_attr_int_new (&klass, (int)underline);
@@ -517,7 +563,7 @@ pango_attr_strikethrough_new (gboolean strikethrough)
     PANGO_ATTR_STRIKETHROUGH,
     pango_attr_int_copy,
     pango_attr_int_destroy,
-    pango_attr_int_compare
+    pango_attr_int_equal
   };
 
   return pango_attr_int_new (&klass, (int)strikethrough);
@@ -540,12 +586,101 @@ pango_attr_rise_new (int rise)
     PANGO_ATTR_RISE,
     pango_attr_int_copy,
     pango_attr_int_destroy,
-    pango_attr_int_compare
+    pango_attr_int_equal
   };
 
   return pango_attr_int_new (&klass, (int)rise);
 }
 
+/**
+ * pango_attr_scale_new:
+ * @scale_factor: factor to scale the font
+ * 
+ * Create a new font size scale attribute. The base font for the
+ * affected text will have its size multiplied by @scale_factor.
+ * 
+ * Return value: the new #PangoAttribute.
+ **/
+PangoAttribute*
+pango_attr_scale_new (double scale_factor)
+{
+  static const PangoAttrClass klass = {
+    PANGO_ATTR_SCALE,
+    pango_attr_float_copy,
+    pango_attr_float_destroy,
+    pango_attr_float_equal
+  };
+
+  return pango_attr_float_new (&klass, scale_factor);
+}
+
+static PangoAttribute *
+pango_attr_shape_copy (const PangoAttribute *attr)
+{
+  const PangoAttrShape *shape_attr = (PangoAttrShape *)attr;
+  
+  return pango_attr_shape_new (&shape_attr->ink_rect, &shape_attr->logical_rect);
+}
+
+static void
+pango_attr_shape_destroy (PangoAttribute *attr)
+{
+  g_free (attr);
+}
+
+static gboolean
+pango_attr_shape_equal (const PangoAttribute *attr1,
+			const PangoAttribute *attr2)
+{
+  const PangoAttrShape *shape_attr1 = (const PangoAttrShape *)attr1;
+  const PangoAttrShape *shape_attr2 = (const PangoAttrShape *)attr2;
+  
+  return (shape_attr1->logical_rect.x == shape_attr2->logical_rect.x &&
+	  shape_attr1->logical_rect.y == shape_attr2->logical_rect.y &&
+	  shape_attr1->logical_rect.width == shape_attr2->logical_rect.width &&
+	  shape_attr1->logical_rect.height == shape_attr2->logical_rect.height &&
+	  shape_attr1->ink_rect.x == shape_attr2->ink_rect.x &&
+	  shape_attr1->ink_rect.y == shape_attr2->ink_rect.y &&
+	  shape_attr1->ink_rect.width == shape_attr2->ink_rect.width &&
+	  shape_attr1->ink_rect.height == shape_attr2->ink_rect.height);
+}
+
+/**
+ * pango_attr_shape_new:
+ * @ink_rect:     ink rectangle to assign to each character
+ * @logical_rect: logical rectangle assign to each character
+ * 
+ * Create a new shape attribute. A shape is used to impose a
+ * particular ink and logical rect on the result of shaping a
+ * particular glyph. This might be used, for instance, for
+ * embedding a picture or a widget inside a PangoLayout.
+ * 
+ * Return value: the newly created attribute
+ **/
+PangoAttribute *
+pango_attr_shape_new (const PangoRectangle *ink_rect,
+		      const PangoRectangle *logical_rect)
+{
+  static const PangoAttrClass klass = {
+    PANGO_ATTR_SHAPE,
+    pango_attr_shape_copy,
+    pango_attr_shape_destroy,
+    pango_attr_shape_equal
+  };
+
+  PangoAttrShape *result;
+
+  g_return_val_if_fail (ink_rect != NULL, NULL);
+  g_return_val_if_fail (logical_rect != NULL, NULL);
+  
+  result = g_new (PangoAttrShape, 1);
+  result->attr.klass = &klass;
+  result->ink_rect = *ink_rect;
+  result->logical_rect = *logical_rect;
+
+  return (PangoAttribute *)result;
+}
+
 /**
  * pango_attr_list_new:
  * 
@@ -561,7 +696,7 @@ pango_attr_list_new (void)
   list->ref_count = 1;
   list->attributes = NULL;
   list->attributes_tail = NULL;
-
+  
   return list;
 }
 
@@ -656,7 +791,7 @@ pango_attr_list_insert_internal (PangoAttrList  *list,
 {
   GSList *tmp_list, *prev, *link;
   gint start_index = attr->start_index;
-  
+
   if (!list->attributes)
     {
       list->attributes = g_slist_prepend (NULL, attr);
@@ -689,6 +824,8 @@ pango_attr_list_insert_internal (PangoAttrList  *list,
 
 	      if (!tmp_list)
 		list->attributes_tail = link;
+
+	      break;
 	    }
 	  
 	  prev = tmp_list;
@@ -762,10 +899,10 @@ pango_attr_list_change (PangoAttrList  *list,
   gint end_index = attr->end_index;
   
   g_return_if_fail (list != NULL);
-
+  
   tmp_list = list->attributes;
   prev = NULL;
-  while (tmp_list)
+  while (1)
     {
       PangoAttribute *tmp_attr;
 
@@ -787,30 +924,71 @@ pango_attr_list_change (PangoAttrList  *list,
 	    list->attributes_tail = link;
 	  
 	  prev = link;
+	  tmp_list = prev->next;
+	  break;
 	}
 
       tmp_attr = tmp_list->data;
 
-      if (tmp_attr->end_index >= start_index &&
-	       pango_attribute_compare (tmp_attr, attr))
+      if (tmp_attr->klass->type == attr->klass->type &&
+	  tmp_attr->end_index >= start_index)
 	{
-	  /* We can merge the new attribute with this attribute
-	   */
-	  end_index = MAX (end_index, tmp_attr->end_index);
-	  tmp_attr->end_index = end_index;
-	  pango_attribute_destroy (attr);
+	  /* We overlap with an existing attribute */
+	  if (pango_attribute_equal (tmp_attr, attr))
+	    {
+	      /* We can merge the new attribute with this attribute
+	       */
+	      if (tmp_attr->end_index >= end_index)
+		{
+		  /* We are totally overlapping the previous attribute.
+		   * No action is needed.
+		   */
+		  pango_attribute_destroy (attr);
+		  return;
+		}
+	      tmp_attr->end_index = end_index;
+	      pango_attribute_destroy (attr);
+	      
+	      attr = tmp_attr;
+	      
+	      prev = tmp_list;
+	      tmp_list = tmp_list->next;
+	      
+	      break;
+	    }
+	  else
+	    {
+	      /* Split, truncate, or remove the old attribute
+	       */
+	      if (tmp_attr->end_index > attr->end_index)
+		{
+		  PangoAttribute *end_attr = pango_attribute_copy (tmp_attr);
 
-	  attr = tmp_attr;
+		  end_attr->start_index = attr->end_index;
+		  pango_attr_list_insert (list, end_attr);
+		}
 
-	  prev = tmp_list;
-	  tmp_list = tmp_list->next;
+	      if (tmp_attr->start_index == attr->start_index)
+		{
+		  pango_attribute_destroy (tmp_attr);
+		  tmp_list->data = attr;
 
-	  break;
+		  prev = tmp_list;
+		  tmp_list = tmp_list->next;
+		  break;
+		}
+	      else
+		{
+		  tmp_attr->end_index = attr->start_index;
+		}
+	    }
 	}
-      
       prev = tmp_list;
       tmp_list = tmp_list->next;
     }
+  /* At this point, prev points to the list node with attr in it,
+   * tmp_list points to prev->next.
+   */
 
   /* We now have the range inserted into the list one way or the
    * other. Fix up the remainder
@@ -821,23 +999,138 @@ pango_attr_list_change (PangoAttrList  *list,
 
       if (tmp_attr->start_index > end_index)
 	break;
-      else if (pango_attribute_compare (tmp_attr, attr))
+      else if (tmp_attr->klass->type == attr->klass->type)
 	{
-	  /* We can merge the new attribute with this attribute
-	   */
-	  attr->end_index = MAX (end_index, tmp_attr->end_index);
-	  
-	  pango_attribute_destroy (tmp_attr);
-	  prev->next = tmp_list->next;
+	  if (tmp_attr->end_index <= attr->end_index ||
+	      pango_attribute_equal (tmp_attr, attr))
+	    {
+	      /* We can merge the new attribute with this attribute.
+	       */
+	      attr->end_index = MAX (end_index, tmp_attr->end_index);
+	      
+	      pango_attribute_destroy (tmp_attr);
+	      prev->next = tmp_list->next;
 
-	  g_slist_free_1 (tmp_list);
-	  tmp_list = prev->next;
+	      if (!prev->next)
+		list->attributes_tail = prev;
+	      
+	      g_slist_free_1 (tmp_list);
+	      tmp_list = prev->next;
+
+	      continue;
+	    }
+	  else
+	    {
+	      /* Trim the start of this attribute that it begins at the end
+	       * of the new attribute. This may involve moving
+	       * it in the list to maintain the required non-decreasing
+	       * order of start indices
+	       */
+	      GSList *tmp_list2;
+	      GSList *prev2;
+	      
+	      tmp_attr->start_index = attr->end_index;
+
+	      tmp_list2 = tmp_list->next;
+	      prev2 = tmp_list;
+
+	      while (tmp_list2)
+		{
+		  PangoAttribute *tmp_attr2 = tmp_list2->data;
+
+		  if (tmp_attr2->start_index >= tmp_attr->start_index)
+		    break;
+
+		  tmp_list2 = tmp_list2->next;
+		  prev2 = tmp_list2;
+		}
+
+	      /* Now remove and insert before tmp_list2. We'll
+	       * hit this attribute again later, but that's harmless.
+	       */
+	      if (prev2 != tmp_list)
+		{
+		  GSList *old_next = tmp_list->next;
+		  
+		  prev->next = old_next;
+		  prev2->next = tmp_list;
+		  tmp_list->next = tmp_list2;
+
+		  if (!tmp_list->next)
+		    list->attributes_tail = tmp_list;
+
+		  tmp_list = old_next;
+
+		  continue;
+		}
+	    }
+	}
+      
+      prev = tmp_list;
+      tmp_list = tmp_list->next;
+    }
+}
+
+/**
+ * pango_attr_list_splice:
+ * @list: a #PangoAttrList
+ * @other: another #PangoAttrList
+ * @pos: the position in @list at which to insert @other
+ * @len: the length of the spliced segment. (Note that this
+ *       must be specified since the attributes in @other
+ *       may only be present at some subsection of this range)
+ * 
+ * This function splices attribute list @other into @list.
+ * This operation is equivalent to stretching every attribute
+ * applies at position @pos in @list by an amount @len,
+ * and then calling pango_attr_list_change() with a copy
+ * of each attributes in @other in sequence (offset in position by @pos).
+ *
+ * This operation proves useful for, for instance, inserting
+ * a preedit string in the middle of an edit buffer.
+ **/
+void
+pango_attr_list_splice (PangoAttrList *list,
+			PangoAttrList *other,
+			gint           pos,
+			gint           len)
+{
+  GSList *tmp_list;
+  
+  g_return_if_fail (list != NULL);
+  g_return_if_fail (other != NULL); 
+  g_return_if_fail (pos >= 0);
+  g_return_if_fail (len >= 0);
+ 
+  tmp_list = list->attributes;
+  while (tmp_list)
+    {
+      PangoAttribute *attr = tmp_list->data;
+
+      if (attr->start_index <= pos)
+	{
+	  if (attr->end_index > pos)
+	    attr->end_index += len;
 	}
       else
 	{
-	  prev = tmp_list;
-	  tmp_list = tmp_list->next;
+	  attr->start_index += len;
+	  attr->end_index += len;
 	}
+
+      tmp_list = tmp_list->next;
+    }
+
+  tmp_list = other->attributes;
+  while (tmp_list)
+    {
+      PangoAttribute *attr = pango_attribute_copy (tmp_list->data);
+      attr->start_index += pos;
+      attr->end_index += pos;
+
+      pango_attr_list_change (list, attr);
+      
+      tmp_list = tmp_list->next;
     }
 }
 
@@ -1056,7 +1349,9 @@ pango_attr_iterator_get_font (PangoAttrIterator     *iterator,
   gboolean have_size = FALSE;
 
   g_return_if_fail (iterator != NULL);
-
+  g_return_if_fail (base != NULL);
+  g_return_if_fail (current != NULL);
+  
   *current = *base;
 
   if (extra_attrs)
@@ -1102,8 +1397,9 @@ pango_attr_iterator_get_font (PangoAttrIterator     *iterator,
 	      have_size = TRUE;
 	      current->size = ((PangoAttrFontDesc *)attr)->desc.size;
 	    }
-	}
-	  
+          }
+          break;
+          
 	case PANGO_ATTR_FAMILY:
 	  if (!have_family)
 	    {
@@ -1146,6 +1442,14 @@ pango_attr_iterator_get_font (PangoAttrIterator     *iterator,
 	      current->size = ((PangoAttrInt *)attr)->value;
 	    }
 	  break;
+        case PANGO_ATTR_SCALE:
+          if (!have_size)
+	    {
+	      have_size = TRUE;
+	      current->size = ((PangoAttrFloat *)attr)->value * base->size;
+	    }
+          break;
+          
 	default:
 	  if (extra_attrs)
 	    {
