@@ -19,9 +19,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <pango/pango-context.h>
-#include <fribidi/fribidi.h>
-#include "iconv.h"
+#include <string.h>
+#include <stdlib.h>
+
+#include "pango/pango-context.h"
+#include "pango/pango-utils.h"
 
 #include "pango-modules.h"
 
@@ -43,9 +45,12 @@ struct _PangoContextClass
 };
 
 static void add_engines (PangoContext      *context,
-			 const gchar       *text, 
+			 const gchar       *text,
+                         gint               start_index,
 			 gint               length,
 			 PangoAttrList     *attrs,
+                         PangoAttrIterator *cached_iter,
+                         gint               n_chars,
 			 PangoEngineShape **shape_engines,
 			 PangoEngineLang  **lang_engines,
 			 PangoFont        **fonts,
@@ -79,7 +84,7 @@ pango_context_get_type (void)
       
       object_type = g_type_register_static (G_TYPE_OBJECT,
                                             "PangoContext",
-                                            &object_info);
+                                            &object_info, 0);
     }
   
   return object_type;
@@ -99,6 +104,7 @@ pango_context_init (PangoContext *context)
   desc.variant = PANGO_VARIANT_NORMAL;
   desc.weight = PANGO_WEIGHT_NORMAL;
   desc.stretch = PANGO_STRETCH_NORMAL;
+  desc.size = 12 * PANGO_SCALE;
 
   context->font_desc = pango_font_description_copy (&desc);
 }
@@ -125,6 +131,8 @@ pango_context_finalize (GObject *object)
   
   g_slist_foreach (context->font_maps, (GFunc)g_object_unref, NULL);
   g_slist_free (context->font_maps);
+
+  pango_font_description_free (context->font_desc);
   
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -494,25 +502,35 @@ pango_context_get_base_dir (PangoContext *context)
  * @context:   a structure holding information that affects
                the itemization process.
  * @text:      the text to itemize.
- * @length:    the number of bytes (not characters) in text.
+ * @start_index: first byte in @text to process
+ * @length:    the number of bytes (not characters) to process
+ *             after @start_index.
  *             This must be >= 0.
  * @attrs:     the set of attributes that apply to @text.
+ * @cached_iter:      Cached attribute iterator, or NULL
  *
  * Breaks a piece of text into segments with consistent
  * directional level and shaping engine.
  *
- * Returns a GList of PangoItem structures.
+ * @cached_iter should be an iterator over @attrs currently positioned at a
+ * range before or containing @start_index; @cached_iter will be advanced to
+ * the range covering the position just after @start_index + @length.
+ * (i.e. if itemizing in a loop, just keep passing in the same @cached_iter).
+ *
+ * Return value: a GList of PangoItem structures.
  */
 GList *
-pango_itemize (PangoContext   *context, 
-	       const char     *text, 
-	       int             length,
-	       PangoAttrList  *attrs)
+pango_itemize (PangoContext      *context, 
+	       const char        *text,
+               int                start_index,
+	       int                length,
+	       PangoAttrList     *attrs,
+               PangoAttrIterator *cached_iter)
 {
   gunichar *text_ucs4;
   int n_chars, i;
   guint8 *embedding_levels;
-  FriBidiCharType base_dir;
+  PangoDirection base_dir;
   PangoItem *item;
   const char *p;
   const char *next;
@@ -524,31 +542,29 @@ pango_itemize (PangoContext   *context,
   PangoFont        **fonts;
 
   g_return_val_if_fail (context != NULL, NULL);
+  g_return_val_if_fail (start_index >= 0, NULL);
   g_return_val_if_fail (length >= 0, NULL);
   g_return_val_if_fail (length == 0 || text != NULL, NULL);
 
   if (length == 0)
     return NULL;
 
-  if (context->base_dir == PANGO_DIRECTION_RTL)
-    base_dir = FRIBIDI_TYPE_RTL;
-  else
-    base_dir = FRIBIDI_TYPE_LTR;
-  
+  base_dir = context->base_dir;
+
   if (length == 0)
     return NULL;
 
   /* First, apply the bidirectional algorithm to break
    * the text into directional runs.
    */
-  text_ucs4 = g_utf8_to_ucs4 (text, length);
+  text_ucs4 = g_utf8_to_ucs4 (text + start_index, length);
   if (!text_ucs4)
     return NULL;
 
-  n_chars = g_utf8_strlen (text, length);
+  n_chars = g_utf8_strlen (text + start_index, length);
   embedding_levels = g_new (guint8, n_chars);
 
-  fribidi_log2vis_get_embedding_levels (text_ucs4, n_chars, &base_dir,
+  pango_log2vis_get_embedding_levels (text_ucs4, n_chars, &base_dir,
 					embedding_levels);
 
   /* Storing these as ranges would be a lot more efficient,
@@ -564,15 +580,17 @@ pango_itemize (PangoContext   *context,
    * each character.
    */
 
-  add_engines (context, text, length, attrs, shape_engines, lang_engines, fonts,
+  add_engines (context, text, start_index, length, attrs,
+               cached_iter,
+               n_chars,
+               shape_engines, lang_engines, fonts,
 	       extra_attr_lists);
 
-
   /* Make a GList of PangoItems out of the above results
    */
 
   item = NULL;
-  p = text;
+  p = text + start_index;
   for (i=0; i<n_chars; i++)
     {
       next = g_utf8_next_char (p);
@@ -599,14 +617,14 @@ pango_itemize (PangoContext   *context,
 	  if (extra_attr_lists[i] && i != 0 && extra_attr_lists[i] == extra_attr_lists[i-1])
 	    {
 	      GSList *tmp_list = extra_attr_lists[i];
-	      item->extra_attrs = NULL;
+	      GSList *new_list = NULL;
 	      while (tmp_list)
 		{
-		  item->extra_attrs = g_slist_prepend (item->extra_attrs,
-						       pango_attribute_copy (tmp_list->data));
+		  new_list = g_slist_prepend (new_list,
+					      pango_attribute_copy (tmp_list->data));
 		  tmp_list = tmp_list->next;
 		}
-	      item->extra_attrs = g_slist_reverse (item->extra_attrs);
+	      item->extra_attrs = g_slist_reverse (new_list);
 	    }
 	  else
 	    item->extra_attrs = extra_attr_lists[i];
@@ -662,12 +680,140 @@ get_font (PangoFont     **fonts,
       
   return result;
 }
-	  
+
+/* FIXME: Remove this artificial limit */
+#define MAX_FAMILIES 16
+
+static void
+load_font (PangoContext          *context,
+	   char                  *lang,
+	   PangoFontDescription  *desc,
+	   PangoFont            **current_fonts,
+	   PangoCoverage        **current_coverages,
+	   int                   *n_families)
+{
+  char **families;
+  char *orig_family;
+  int j;
+  
+  for (j=0; j < *n_families; j++)
+    {
+      if (current_fonts[j])
+	{
+	  g_object_unref (G_OBJECT (current_fonts[j]));
+	  pango_coverage_unref (current_coverages[j]);
+	}
+    }
+
+  orig_family = desc->family_name;
+  families = g_strsplit (orig_family, ",", -1);
+
+  *n_families = 0;
+  for (j=0; families[j] && *n_families < MAX_FAMILIES; j++)
+    {
+      desc->family_name = families[j];
+      current_fonts[*n_families] = pango_context_load_font (context, desc);
+      
+      if (current_fonts[*n_families])
+	{
+	  current_coverages[*n_families] = pango_font_get_coverage (current_fonts[*n_families], lang);
+	  (*n_families)++;
+	}
+    }
+  
+  g_strfreev (families);
+
+  /* The font description was completely unloadable, try with
+   * family == "Sans"
+   */
+  if (*n_families == 0)
+    {
+      char *ctmp1, *ctmp2;
+      
+      desc->family_name = orig_family;
+      
+      ctmp1 = pango_font_description_to_string (desc);
+      desc->family_name = "Sans";
+      ctmp2 = pango_font_description_to_string (desc);
+      
+      g_warning ("Couldn't load font \"%s\" falling back to \"%s\"", ctmp1, ctmp2);
+      g_free (ctmp1);
+      g_free (ctmp2);
+      
+      desc->family_name = "Sans";
+      
+      current_fonts[0] = pango_context_load_font (context, desc);
+      if (current_fonts[0])
+	{
+	  current_coverages[0] = pango_font_get_coverage (current_fonts[0], lang);
+	  *n_families = 1;
+	}
+    }
+  
+  /* We couldn't try with Sans and the specified style. Try Sans Normal
+   */
+  if (*n_families == 0)
+    {
+      char *ctmp1, *ctmp2;
+      
+      ctmp1 = pango_font_description_to_string (desc);
+      desc->style = PANGO_STYLE_NORMAL;
+      desc->weight = PANGO_WEIGHT_NORMAL;
+      desc->variant = PANGO_VARIANT_NORMAL;
+      desc->stretch = PANGO_STRETCH_NORMAL;
+      ctmp2 = pango_font_description_to_string (desc);
+      
+      g_warning ("Couldn't load font \"%s\" falling back to \"%s\"", ctmp1, ctmp2);
+      g_free (ctmp1);
+      g_free (ctmp2);
+      
+      current_fonts[0] = pango_context_load_font (context, desc);
+      if (current_fonts[0])
+	{
+	  current_coverages[0] = pango_font_get_coverage (current_fonts[0], lang);
+	  *n_families = 1;
+	}
+    }
+
+  /* Everything failed, we are screwed, there is no way to continue
+   */
+  if (n_families == 0)
+    {
+      g_warning ("All font failbacks failed!!!!");
+      exit (1);
+    }
+}
+
+static gboolean
+advance_iterator_to (PangoAttrIterator *iterator,
+                     int                start_index)
+{
+  int start_range, end_range;
+  
+  pango_attr_iterator_range (iterator, &start_range, &end_range);
+
+  while (start_index >= end_range)
+    {
+      if (!pango_attr_iterator_next (iterator))
+        return FALSE;
+      pango_attr_iterator_range (iterator, &start_range, &end_range);
+    }
+
+  if (start_range > start_index)
+    g_warning ("In pango_itemize(), the cached iterator passed in "
+               "had already moved beyond the start_index");
+
+  return TRUE;
+}
+
 static void
 add_engines (PangoContext      *context,
-	     const gchar       *text, 
+	     const gchar       *text,
+             gint               start_index,
 	     gint               length,
 	     PangoAttrList     *attrs,
+             PangoAttrIterator *cached_iter,
+             gint               n_chars,
 	     PangoEngineShape **shape_engines,
 	     PangoEngineLang  **lang_engines,
 	     PangoFont        **fonts,
@@ -675,37 +821,47 @@ add_engines (PangoContext      *context,
 {
   const char *pos;
   char *lang = NULL;
-  int next_index = 0;
+  int next_index;
   GSList *extra_attrs = NULL;
-  gint n_chars;
   PangoMap *lang_map = NULL;
   PangoFontDescription current_desc = { 0 };
-
-  /* FIXME: Remove this artificial limit */
-
   int n_families = 0;
-#define MAX_FAMILIES 16
   PangoFont *current_fonts[MAX_FAMILIES];
-  PangoCoverage *current_coverages[MAX_FAMILIES];
-  
+  PangoCoverage *current_coverages[MAX_FAMILIES];  
   PangoAttrIterator *iterator;
   PangoAttribute *attr;
-  
+  gboolean first_iteration = TRUE;
   gunichar wc;
-  int i, j;
+  int i = 0, j;
 
-  n_chars = g_utf8_strlen (text, length);
+  if (cached_iter)
+    iterator = cached_iter;
+  else
+    iterator = pango_attr_list_get_iterator (attrs);
 
-  iterator = pango_attr_list_get_iterator (attrs);
-
-  pos = text;
+  advance_iterator_to (iterator, start_index);
+  
+  pango_attr_iterator_range (iterator, NULL, &next_index);
+  
+  pos = text + start_index;
   for (i=0; i<n_chars; i++)
     {
-      if (pos - text == next_index)
+      if (first_iteration || pos - text == next_index)
 	{
 	  char *next_lang;
 	  PangoFontDescription next_desc;
-	  
+
+          first_iteration = FALSE;
+          
+          /* Only advance the iterator if we've exhausted a range,
+           * not on the first iteration.
+           */
+          if (pos - text == next_index)
+            {
+              pango_attr_iterator_next (iterator);
+              pango_attr_iterator_range (iterator, NULL, &next_index);
+            }
+          
 	  attr = pango_attr_iterator_get (iterator, PANGO_ATTR_LANG);
 	  if (attr)
 	    next_lang = ((PangoAttrString *)attr)->value;
@@ -728,50 +884,21 @@ add_engines (PangoContext      *context,
 		}
 	       
 	      lang_map = pango_find_map (next_lang,
-					  engine_type_id, render_type_id);
+					 engine_type_id, render_type_id);
 	    }
 
-	  pango_attr_iterator_get_font (iterator, context->font_desc, &next_desc, &extra_attrs);
-
+	  pango_attr_iterator_get_font (iterator, context->font_desc,
+                                        &next_desc, &extra_attrs);
+          
 	  if (i == 0 ||
-	      !pango_font_description_compare (&current_desc, &next_desc))
+	      !pango_font_description_equal (&current_desc, &next_desc))
 	    {
-	      char **families;
-	      
 	      current_desc = next_desc;
 
-	      for (j=0; j<n_families; j++)
-		{
-		  if (current_fonts[j])
-		    {
-		      g_object_unref (G_OBJECT (current_fonts[j]));
-		      pango_coverage_unref (current_coverages[j]);
-		    }
-		}
-
-	      families = g_strsplit (current_desc.family_name, ",", -1);
-	      n_families = 0;
-	      while (families[n_families])
-		n_families++;
-
-	      if (n_families > MAX_FAMILIES)
-		n_families = MAX_FAMILIES;
-
-	      for (j=0; j<n_families; j++)
-		{
-		  next_desc.family_name = families[j];
-		  current_fonts[j] = pango_context_load_font (context, &next_desc);
-
-		  if (current_fonts[j])
-		    current_coverages[j] = pango_font_get_coverage (current_fonts[j], lang);
-		}
-
-	      g_strfreev (families);
+	      load_font (context, lang, &current_desc,
+			 current_fonts, current_coverages, &n_families);
 	    }
-
-	  pango_attr_iterator_range (iterator, NULL, &next_index);
-	  pango_attr_iterator_next (iterator);
-	}
+        }
 
       wc = g_utf8_get_char (pos);
       pos = g_utf8_next_char (pos);
@@ -787,6 +914,8 @@ add_engines (PangoContext      *context,
 
       extra_attr_lists[i] = extra_attrs;
     }
+
+  g_assert (pos - text == start_index + length);
   
   for (j=0; j<n_families; j++)
     {
@@ -797,6 +926,7 @@ add_engines (PangoContext      *context,
 	}
     }
 
-  pango_attr_iterator_destroy (iterator);
+  if (iterator != cached_iter)
+    pango_attr_iterator_destroy (iterator);
 }
 
