@@ -25,6 +25,7 @@
 #include <pango/pango-attributes.h>
 #include <pango/pango-context.h>
 #include <pango/pango-glyph.h>
+#include <pango/pango-tabs.h>
 
 #ifdef __cplusplus
 extern "C" {
@@ -33,7 +34,7 @@ extern "C" {
 typedef struct _PangoLayout      PangoLayout;
 typedef struct _PangoLayoutClass PangoLayoutClass;
 typedef struct _PangoLayoutLine  PangoLayoutLine;
-typedef struct _PangoLayoutRun   PangoLayoutRun; 
+typedef struct _PangoLayoutRun   PangoLayoutRun;
 
 typedef enum {
   PANGO_ALIGN_LEFT,
@@ -66,16 +67,29 @@ struct _PangoLayoutRun
  * need to create a subclass of these, mail otaylor@redhat.com
  */
 
-GType        pango_layout_get_type       (void);
+GType        pango_layout_get_type       (void) G_GNUC_CONST;
 PangoLayout *pango_layout_new            (PangoContext   *context);
 
 PangoContext  *pango_layout_get_context    (PangoLayout    *layout);
 
 void           pango_layout_set_attributes (PangoLayout    *layout,
 					    PangoAttrList  *attrs);
+PangoAttrList *pango_layout_get_attributes (PangoLayout    *layout);
+
 void           pango_layout_set_text       (PangoLayout    *layout,
 					    const char     *text,
 					    int             length);
+const char    *pango_layout_get_text       (PangoLayout    *layout);
+
+void           pango_layout_set_markup     (PangoLayout    *layout,
+                                            const char     *markup,
+                                            int             length);
+
+void           pango_layout_set_markup_with_accel (PangoLayout    *layout,
+                                                   const char     *markup,
+                                                   int             length,
+                                                   gunichar        accel_marker,
+                                                   gunichar       *accel_char);
 
 void           pango_layout_set_font_description (PangoLayout                *layout,
 						  const PangoFontDescription *desc);
@@ -95,6 +109,11 @@ void           pango_layout_set_alignment        (PangoLayout                *la
 						  PangoAlignment              alignment);
 PangoAlignment pango_layout_get_alignment        (PangoLayout                *layout);
 
+void           pango_layout_set_tabs             (PangoLayout                *layout,
+                                                  PangoTabArray              *tabs);
+
+PangoTabArray* pango_layout_get_tabs             (PangoLayout                *layout);
+
 void           pango_layout_context_changed (PangoLayout    *layout);
 
 void     pango_layout_get_log_attrs (PangoLayout    *layout,
@@ -159,6 +178,43 @@ void     pango_layout_line_get_pixel_extents (PangoLayoutLine *layout_line,
 					      PangoRectangle  *ink_rect,
 					      PangoRectangle  *logical_rect);
 
+typedef struct _PangoLayoutIter PangoLayoutIter;
+
+PangoLayoutIter *pango_layout_get_iter  (PangoLayout     *layout);
+void             pango_layout_iter_free (PangoLayoutIter *iter);
+
+int              pango_layout_iter_get_index (PangoLayoutIter *iter);
+PangoLayoutRun  *pango_layout_iter_get_run   (PangoLayoutIter *iter);
+PangoLayoutLine *pango_layout_iter_get_line  (PangoLayoutIter *iter);
+gboolean         pango_layout_iter_at_last_line (PangoLayoutIter *iter);
+
+gboolean pango_layout_iter_next_char    (PangoLayoutIter *iter);
+gboolean pango_layout_iter_next_cluster (PangoLayoutIter *iter);
+gboolean pango_layout_iter_next_run     (PangoLayoutIter *iter);
+gboolean pango_layout_iter_next_line    (PangoLayoutIter *iter);
+
+void pango_layout_iter_get_char_extents    (PangoLayoutIter *iter,
+                                            PangoRectangle  *logical_rect);
+void pango_layout_iter_get_cluster_extents (PangoLayoutIter *iter,
+                                            PangoRectangle  *ink_rect,
+                                            PangoRectangle  *logical_rect);
+void pango_layout_iter_get_run_extents     (PangoLayoutIter *iter,
+                                            PangoRectangle  *ink_rect,
+                                            PangoRectangle  *logical_rect);
+void pango_layout_iter_get_line_extents    (PangoLayoutIter *iter,
+                                            PangoRectangle  *ink_rect,
+                                            PangoRectangle  *logical_rect);
+/* All the yranges meet, unlike the logical_rect's (i.e. the yranges
+ * assign between-line spacing to the nearest line)
+ */
+void pango_layout_iter_get_line_yrange     (PangoLayoutIter *iter,
+                                            int             *y0,
+                                            int             *y1);
+void pango_layout_iter_get_layout_extents  (PangoLayoutIter *iter,
+                                            PangoRectangle  *ink_rect,
+                                            PangoRectangle  *logical_rect);
+int  pango_layout_iter_get_baseline        (PangoLayoutIter *iter);
+
 #ifdef __cplusplus
 }
 #endif /* __cplusplus */
