@@ -24,6 +24,9 @@
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
+#include <glib.h>
+
+#include <X11/Xatom.h>
 
 #include "pango-fontmap.h"
 #include "pango-utils.h"
@@ -81,9 +84,14 @@ struct _PangoXFontMap
   GHashTable *families;
   GHashTable *size_infos;
 
+  GHashTable *to_atom_cache;
+  GHashTable *from_atom_cache;
+
   int n_fonts;
 
   double resolution;		/* (points / pixel) * PANGO_SCALE */
+
+  Window coverage_win;
 };
 
 struct _PangoXFontMapClass
@@ -175,6 +183,7 @@ static char *   pango_x_get_xlfd_field      (const char         *fontname,
 					     char               *buffer);
 static char *   pango_x_get_identifier      (const char         *fontname);
 
+
 static PangoFontClass *parent_class;	/* Parent class structure for PangoXFontMap */
 
 static GType
@@ -199,7 +208,7 @@ pango_x_font_map_get_type (void)
       
       object_type = g_type_register_static (PANGO_TYPE_FONT_MAP,
                                             "PangoXFontMap",
-                                            &object_info);
+                                            &object_info, 0);
     }
   
   return object_type;
@@ -210,6 +219,8 @@ pango_x_font_map_init (PangoXFontMap *xfontmap)
 {
   xfontmap->families = g_hash_table_new (g_str_hash, g_str_equal);
   xfontmap->size_infos = g_hash_table_new (g_str_hash, g_str_equal);
+  xfontmap->to_atom_cache = g_hash_table_new (g_str_hash, g_str_equal);
+  xfontmap->from_atom_cache = g_hash_table_new (g_direct_hash, g_direct_equal);
   xfontmap->n_fonts = 0;
 }
 
@@ -488,22 +499,37 @@ pango_x_font_map_load_font (PangoFontMap               *fontmap,
   if (family_entry)
     {
       PangoXFontEntry *best_match = NULL;
+      int best_distance = G_MAXINT;
       
       tmp_list = family_entry->font_entries;
       while (tmp_list)
 	{
 	  PangoXFontEntry *font_entry = tmp_list->data;
 	  
-	  if (font_entry->description.style == description->style &&
-	      font_entry->description.variant == description->variant &&
+	  if (font_entry->description.variant == description->variant &&
 	      font_entry->description.stretch == description->stretch)
 	    {
-	      int distance = abs(font_entry->description.weight - description->weight);
-	      int old_distance = best_match ? abs(best_match->description.weight - description->weight) : G_MAXINT;
-
-	      if (distance < old_distance)
+	      if (font_entry->description.style == description->style)
 		{
-		  best_match = font_entry;
+		  int distance = abs(font_entry->description.weight - description->weight);
+		  if (distance < best_distance)
+		    {
+		      best_match = font_entry;
+		      best_distance = distance;
+		    }
+		}
+	      else if (font_entry->description.style != PANGO_STYLE_NORMAL &&
+		       description->style != PANGO_STYLE_NORMAL)
+		{
+		  /* Equate oblique and italic, but with a big penalty
+		   */
+		  int distance = PANGO_SCALE * 1000 + abs(font_entry->description.weight - description->weight);
+
+		  if (distance < best_distance)
+		    {
+		      best_match = font_entry;
+		      best_distance = distance;
+		    }
 		}
 	    }
 
@@ -547,201 +573,269 @@ pango_x_font_map_load_font (PangoFontMap               *fontmap,
   return result;
 }
 
-static gboolean
-get_style (GString *str, PangoFontDescription *desc)
-{
-  if (str->len == 0)
-    return FALSE;
+
+/************************
+ * Coverage Map Caching *
+ ************************/
 
-  switch (str->str[0])
-    {
-    case 'n':
-    case 'N':
-      if (strncasecmp (str->str, "normal", str->len) == 0)
-	{
-	  desc->style = PANGO_STYLE_NORMAL;
-	  return TRUE;
-	}
-      break;
-    case 'i':
-      if (strncasecmp (str->str, "italic", str->len) == 0)
-	{
-	  desc->style = PANGO_STYLE_ITALIC;
-	  return TRUE;
-	}
-      break;
-    case 'o':
-      if (strncasecmp (str->str, "oblique", str->len) == 0)
-	{
-	  desc->style = PANGO_STYLE_OBLIQUE;
-	  return TRUE;
-	}
-      break;
-    }
-  g_warning ("Style must be normal, italic, or oblique");
-  
-  return FALSE;
+/* We need to be robust against errors accessing the coverage
+ * cache window, since it is not our window. So we temporarily
+ * set this error handler while accessing it. The error_occured
+ * global allows us to tell whether an error occured for
+ * XChangeProperty
+ */
+static gboolean error_occured;
+
+static int 
+ignore_error (Display *d, XErrorEvent *e)
+{
+  return 0;
 }
 
-static gboolean
-get_variant (GString *str, PangoFontDescription *desc)
+/* Retrieve the coverage window for the given display.
+ * We look for a property on the root window, and then
+ * check that the window that property points to also
+ * has the same property pointing to itself. The second
+ * check allows us to make sure that a stale property
+ * isn't just pointing to some other apps window
+ */
+static Window
+pango_x_real_get_coverage_win (Display *display)
 {
-  if (str->len == 0)
-    return FALSE;
-
-  switch (str->str[0])
+  Atom type;
+  int format;
+  gulong n_items;
+  gulong bytes_after;
+  Atom *data;
+  Window retval = None;
+  int (*old_handler) (Display *, XErrorEvent *);
+  
+  Atom coverage_win_atom = XInternAtom (display,
+					"PANGO_COVERAGE_WIN",
+					False);
+  
+  XGetWindowProperty (display,
+		      DefaultRootWindow (display),
+		      coverage_win_atom,
+		      0, 4,
+		      False, XA_WINDOW,
+		      &type, &format, &n_items, &bytes_after,
+		      (guchar **)&data);
+  
+  if (type == XA_WINDOW)
     {
-    case 'n':
-    case 'N':
-      if (strncasecmp (str->str, "normal", str->len) == 0)
+      if (format == 32 && n_items == 1 && bytes_after == 0)
+	retval = *data;
+      
+      XFree (data);
+    }
+
+  old_handler= XSetErrorHandler (ignore_error);
+
+  if (XGetWindowProperty (display,
+			  retval,
+			  coverage_win_atom,
+			  0, 4,
+			  False, XA_WINDOW,
+			  &type, &format, &n_items, &bytes_after,
+			  (guchar **)&data) == Success &&
+      type == XA_WINDOW)
+    {
+      if (format != 32 || n_items != 1 || bytes_after != 0 ||
+	  *data != retval)
+	retval = None;
+      
+      XFree (data);
+    }
+  else
+    retval = None;
+
+  XSync (display, False);
+  XSetErrorHandler (old_handler);
+
+  return retval;
+}
+
+/* Find or create the peristant window for caching font coverage
+ * information.
+ *
+ * To assure atomic creation, we first look for the window, then if we
+ * don't find it, grab the server, look for it again, and then if that
+ * still didn't find it, create it and ungrab.
+ */
+static Window
+pango_x_get_coverage_win (PangoXFontMap *xfontmap)
+{
+  if (!xfontmap->coverage_win)
+    xfontmap->coverage_win = pango_x_real_get_coverage_win (xfontmap->display);
+
+  if (!xfontmap->coverage_win)
+    {
+      Display *persistant_display;
+
+      persistant_display = XOpenDisplay (DisplayString (xfontmap->display));
+      if (!persistant_display)
 	{
-	  desc->variant = PANGO_VARIANT_NORMAL;
-	  return TRUE;
+	  g_warning ("Cannot create or retrieve display for font coverage cache");
+	  return None;
 	}
-      break;
-    case 's':
-    case 'S':
-      if (strncasecmp (str->str, "small_caps", str->len) == 0)
+
+      XGrabServer (persistant_display);
+
+      xfontmap->coverage_win = pango_x_real_get_coverage_win (xfontmap->display);
+      if (!xfontmap->coverage_win)
 	{
-	  desc->variant = PANGO_VARIANT_SMALL_CAPS;
-	  return TRUE;
+	  XSetWindowAttributes attr;
+	  
+	  attr.override_redirect = True;
+	  
+	  XSetCloseDownMode (persistant_display, RetainPermanent);
+	  
+	  xfontmap->coverage_win = 
+	    XCreateWindow (persistant_display,
+			   DefaultRootWindow (persistant_display),
+			   -100, -100, 10, 10, 0, 0,
+			   InputOnly, CopyFromParent,
+			   CWOverrideRedirect, &attr);
+	  
+	  XChangeProperty (persistant_display,
+			   DefaultRootWindow (persistant_display),
+			   XInternAtom (persistant_display,
+					"PANGO_COVERAGE_WIN",
+					FALSE),
+			   XA_WINDOW,
+			   32, PropModeReplace,
+			   (guchar *)&xfontmap->coverage_win, 1);
+	  
+
+	  XChangeProperty (persistant_display,
+			   xfontmap->coverage_win,
+			   XInternAtom (persistant_display,
+					"PANGO_COVERAGE_WIN",
+					FALSE),
+			   XA_WINDOW,
+			   32, PropModeReplace,
+			   (guchar *)&xfontmap->coverage_win, 1);
 	}
-      break;
+      
+      XUngrabServer (persistant_display);
+      
+      XSync (persistant_display, False);
+      XCloseDisplay (persistant_display);
+    }
+
+  return xfontmap->coverage_win;
+}
+
+/* Find the cached value for the coverage map on the
+ * coverage cache window, if it exists. *atom is set
+ * to the interned value of str for later use in storing
+ * the property if the lookup fails
+ */
+static PangoCoverage *
+pango_x_get_cached_coverage (PangoXFontMap *xfontmap,
+			     const char    *str,
+			     Atom          *atom)
+{
+  int (*old_handler) (Display *, XErrorEvent *);
+  Window coverage_win;
+  PangoCoverage *result = NULL;
+
+  Atom type;
+  int format;
+  int tries = 5;
+  gulong n_items;
+  gulong bytes_after;
+  guchar *data;
+
+  *atom = XInternAtom (xfontmap->display, str, False);
+      
+  while (tries--)
+    {
+      coverage_win = pango_x_get_coverage_win (xfontmap);
+      
+      if (!coverage_win)
+	return NULL;
+      
+      old_handler= XSetErrorHandler (ignore_error);
+      
+      if (XGetWindowProperty (xfontmap->display,
+			      coverage_win, *atom,
+			      0, G_MAXLONG,
+			      False, XA_STRING,
+			      &type, &format, &n_items, &bytes_after,
+			      &data) == Success
+	  && type == XA_STRING)
+	{
+	  if (format == 8 && bytes_after == 0)
+	    result = pango_coverage_from_bytes (data, n_items);
+	  
+	  XSetErrorHandler (old_handler);
+	  XFree (data);
+	  break;
+	}
+      else
+	{
+	  /* Window disappeared out from under us */
+	  XSetErrorHandler (old_handler);
+	  xfontmap->coverage_win = None;
+	}
+
+    }
+      
+  return result;
+}
+
+/* Store the given coverage map on the coverage cache window.
+ * atom is the intern'ed value of the string that identifies
+ * the cache entry.
+ */
+static void
+pango_x_store_cached_coverage (PangoXFontMap *xfontmap,
+			       Atom           atom,
+			       PangoCoverage *coverage)
+{
+  int (*old_handler) (Display *, XErrorEvent *);
+  guchar *bytes;
+  gint size;
+
+  int tries = 5;
+  
+  pango_coverage_to_bytes (coverage, &bytes, &size);
+
+  while (tries--)
+    {
+      Window coverage_win = pango_x_get_coverage_win (xfontmap);
+
+      if (!coverage_win)
+	break;
+
+      old_handler = XSetErrorHandler (ignore_error);
+      error_occured = False;
+  
+      XChangeProperty (xfontmap->display,
+		       coverage_win,
+		       atom, 
+		       XA_STRING,
+		       8, PropModeReplace,
+		       bytes, size);
+      
+      XSync (xfontmap->display, False);
+      XSetErrorHandler (old_handler);
+
+      if (!error_occured)
+	break;
+      else
+	{
+	  /* Window disappeared out from under us */
+	  XSetErrorHandler (old_handler);
+	  xfontmap->coverage_win = None;
+	}	  
     }
   
-  g_warning ("Variant must be normal, or small_caps");
-  return FALSE;
-}
-
-static gboolean
-get_weight (GString *str, PangoFontDescription *desc)
-{
-  if (str->len == 0)
-    return FALSE;
-
-  switch (str->str[0])
-    {
-    case 'n':
-    case 'N':
-      if (strncasecmp (str->str, "normal", str->len) == 0)
-	{
-	  desc->weight = PANGO_WEIGHT_NORMAL;
-	  return TRUE;
-	}
-      break;
-    case 'b':
-    case 'B':
-      if (strncasecmp (str->str, "bold", str->len) == 0)
-	{
-	  desc->weight = PANGO_WEIGHT_BOLD;
-	  return TRUE;
-	}
-      break;
-    case '0':
-    case '1':
-    case '2':
-    case '3':
-    case '4':
-    case '5':
-    case '6':
-    case '7':
-    case '8':
-    case '9':
-      {
-	char *numstr, *end;
-
-	numstr = g_strndup (str->str, str->len);
-
-	desc->weight = strtol (numstr, &end, 0);
-	if (*end != '\0')
-	  {
-	    g_warning ("Cannot parse numerical weight '%s'", numstr);
-	    g_free (numstr);
-	    return FALSE;
-	  }
-
-	g_free (numstr);
-	return TRUE;
-      }
-    }
-  
-  g_warning ("Weight must be normal, bold, or an integer");
-  return FALSE;
-}
-
-static gboolean
-get_stretch (GString *str, PangoFontDescription *desc)
-{
-  if (str->len == 0)
-    return FALSE;
-
-  switch (str->str[0])
-    { 
-    case 'c':
-    case 'C':
-      if (strncasecmp (str->str, "condensed", str->len) == 0)
-	{
-	  desc->stretch = PANGO_STRETCH_CONDENSED;
-	  return TRUE;
-	}
-      break;
-    case 'e':
-    case 'E':
-      if (strncasecmp (str->str, "extra_condensed", str->len) == 0)
-	{
-	  desc->stretch = PANGO_STRETCH_EXTRA_CONDENSED;
-	  return TRUE;
-	}
-     if (strncasecmp (str->str, "extra_expanded", str->len) == 0)
-	{
-	  desc->stretch = PANGO_STRETCH_EXTRA_EXPANDED;
-	  return TRUE;
-	}
-      if (strncasecmp (str->str, "expanded", str->len) == 0)
-	{
-	  desc->stretch = PANGO_STRETCH_EXPANDED;
-	  return TRUE;
-	}
-      break;
-    case 'n':
-    case 'N':
-      if (strncasecmp (str->str, "normal", str->len) == 0)
-	{
-	  desc->stretch = PANGO_STRETCH_NORMAL;
-	  return TRUE;
-	}
-      break;
-    case 's':
-    case 'S':
-      if (strncasecmp (str->str, "semi_condensed", str->len) == 0)
-	{
-	  desc->stretch = PANGO_STRETCH_SEMI_CONDENSED;
-	  return TRUE;
-	}
-      if (strncasecmp (str->str, "semi_expanded", str->len) == 0)
-	{
-	  desc->stretch = PANGO_STRETCH_SEMI_EXPANDED;
-	  return TRUE;
-	}
-      break;
-    case 'u':
-    case 'U':
-      if (strncasecmp (str->str, "ultra_condensed", str->len) == 0)
-	{
-	  desc->stretch = PANGO_STRETCH_ULTRA_CONDENSED;
-	  return TRUE;
-	}
-      if (strncasecmp (str->str, "ultra_expanded", str->len) == 0)
-	{
-	  desc->variant = PANGO_STRETCH_ULTRA_EXPANDED;
-	  return TRUE;
-	}
-      break;
-    }
-
-  g_warning ("Stretch must be ultra_condensed, extra_condensed, condensed, semi_condensed, normal, semi_expanded, expanded, extra_expanded, or ultra_expanded");
-  return FALSE;
+  g_free (bytes);
 }
+
 
 static void
 pango_x_font_map_read_alias_file (PangoXFontMap *xfontmap,
@@ -758,14 +852,15 @@ pango_x_font_map_read_alias_file (PangoXFontMap *xfontmap,
     {
       GString *line_buf = g_string_new (NULL);
       GString *tmp_buf = g_string_new (NULL);
+      gint lines_read;
 
-      while (pango_read_line (infile, line_buf))
+      while ((lines_read = pango_read_line (infile, line_buf)))
 	{
 	  PangoXFamilyEntry *family_entry;
 
 	  const char *p = line_buf->str;
 	  
-	  lineno++;
+	  lineno += lines_read;
 
 	  if (!pango_skip_space (&p))
 	    continue;
@@ -781,25 +876,25 @@ pango_x_font_map_read_alias_file (PangoXFontMap *xfontmap,
 	  if (!pango_scan_string (&p, tmp_buf))
 	    goto error;
 
-	  if (!get_style (tmp_buf, &font_entry->description))
+	  if (!pango_parse_style (tmp_buf->str, &font_entry->description, TRUE))
 	    goto error;
 
 	  if (!pango_scan_string (&p, tmp_buf))
 	    goto error;
 
-	  if (!get_variant (tmp_buf, &font_entry->description))
+	  if (!pango_parse_variant (tmp_buf->str, &font_entry->description, TRUE))
 	    goto error;
 
 	  if (!pango_scan_string (&p, tmp_buf))
 	    goto error;
 
-	  if (!get_weight (tmp_buf, &font_entry->description))
+	  if (!pango_parse_weight (tmp_buf->str, &font_entry->description, TRUE))
 	    goto error;
 	  
 	  if (!pango_scan_string (&p, tmp_buf))
 	    goto error;
 
-	  if (!get_stretch (tmp_buf, &font_entry->description))
+	  if (!pango_parse_stretch (tmp_buf->str, &font_entry->description, TRUE))
 	    goto error;
 
 	  if (!pango_scan_string (&p, tmp_buf))
@@ -1286,39 +1381,57 @@ pango_x_font_entry_get_coverage (PangoXFontEntry *entry,
 				 PangoFont       *font,
 				 const char      *lang)
 {
-  guint32 ch;
-  PangoMap *shape_map;
-  PangoCoverage *coverage;
-  PangoCoverage *result;
-  PangoCoverageLevel font_level;
-  PangoMapEntry *map_entry;
+  PangoXFont *xfont;
+  PangoXFontMap *xfontmap = NULL; /* Quiet gcc */
+  PangoCoverage *result = NULL;
   GHashTable *coverage_hash;
+  Atom atom = None;
 
   if (entry)
-    if (entry->coverage)
-      {
-	pango_coverage_ref (entry->coverage);
-	return entry->coverage;
-      }
-
-  result = pango_coverage_new ();
-
-  coverage_hash = g_hash_table_new (g_str_hash, g_str_equal);
-  
-  shape_map = pango_x_get_shaper_map (lang);
-
-  for (ch = 0; ch < 65536; ch++)
     {
-      map_entry = pango_map_get_entry (shape_map, ch);
-      if (map_entry->info)
+      if (entry->coverage)
 	{
-	  coverage = g_hash_table_lookup (coverage_hash, map_entry->info->id);
-	  if (!coverage)
+	  pango_coverage_ref (entry->coverage);
+	  return entry->coverage;
+	}
+      
+      xfont = (PangoXFont *)font;
+      
+      xfontmap = (PangoXFontMap *)pango_x_font_map_for_display (xfont->display);
+      if (entry->xlfd)
+	{
+	  char *str = g_strconcat (lang ? lang : "*", "|", entry->xlfd, NULL);
+	  result = pango_x_get_cached_coverage (xfontmap, str, &atom);
+	  g_free (str);
+	}
+    }
+
+  if (!result)
+    {
+      guint32 ch;
+      PangoMap *shape_map;
+      PangoCoverage *coverage;
+      PangoCoverageLevel font_level;
+      PangoMapEntry *map_entry;
+      
+      result = pango_coverage_new ();
+      
+      coverage_hash = g_hash_table_new (g_str_hash, g_str_equal);
+      
+      shape_map = pango_x_get_shaper_map (lang);
+      
+      for (ch = 0; ch < 65536; ch++)
+	{
+	  map_entry = pango_map_get_entry (shape_map, ch);
+	  if (map_entry->info)
 	    {
-	      PangoEngineShape *engine = (PangoEngineShape *)pango_map_get_engine (shape_map, ch);
-	      coverage = engine->get_coverage (font, lang);
-	      g_hash_table_insert (coverage_hash, map_entry->info->id, coverage);
-	    }
+	      coverage = g_hash_table_lookup (coverage_hash, map_entry->info->id);
+	      if (!coverage)
+		{
+		  PangoEngineShape *engine = (PangoEngineShape *)pango_map_get_engine (shape_map, ch);
+		  coverage = engine->get_coverage (font, lang);
+		  g_hash_table_insert (coverage_hash, map_entry->info->id, coverage);
+		}
 	  
 	  font_level = pango_coverage_get (coverage, ch);
 	  if (font_level == PANGO_COVERAGE_EXACT && !map_entry->is_exact)
@@ -1326,18 +1439,22 @@ pango_x_font_entry_get_coverage (PangoXFontEntry *entry,
 
 	  if (font_level != PANGO_COVERAGE_NONE)
 	    pango_coverage_set (result, ch, font_level);
+	    }
 	}
+      
+      g_hash_table_foreach (coverage_hash, free_coverages_foreach, NULL);
+      g_hash_table_destroy (coverage_hash);
+
+      if (atom)
+	pango_x_store_cached_coverage (xfontmap, atom, result);
     }
-
-  g_hash_table_foreach (coverage_hash, free_coverages_foreach, NULL);
-  g_hash_table_destroy (coverage_hash);
-
+      
   if (entry)
     {
       entry->coverage = result;
       pango_coverage_ref (result);
     }
-
+      
   return result;
 }
 
@@ -1414,3 +1531,46 @@ pango_x_fontmap_cache_clear (PangoXFontMap   *xfontmap)
   xfontmap->freed_fonts->length = 0;
 }
 
+
+Atom
+pango_x_fontmap_atom_from_name (PangoFontMap *fontmap, 
+				const char   *atomname)
+{
+  PangoXFontMap *xfm = PANGO_X_FONT_MAP(fontmap);
+  gpointer found;
+  Atom atom;
+
+  found = g_hash_table_lookup (xfm->to_atom_cache, atomname);
+
+  if (found) 
+    return (Atom)(GPOINTER_TO_UINT(found));
+
+  atom = XInternAtom (xfm->display, atomname, FALSE);
+  g_hash_table_insert (xfm->to_atom_cache, g_strdup (atomname), 
+		       (gpointer)atom);
+
+  return atom;
+}
+
+
+const char *
+pango_x_fontmap_name_from_atom (PangoFontMap *fontmap, 
+				Atom          atom)
+{
+  PangoXFontMap *xfm = PANGO_X_FONT_MAP(fontmap);
+  gpointer found;
+  char *name, *name2;
+
+  found = g_hash_table_lookup (xfm->from_atom_cache, GUINT_TO_POINTER(atom));
+
+  if (found) 
+    return (const char *)found;
+
+  name = XGetAtomName (xfm->display, atom);
+  name2 = g_strdup (name);
+  XFree (name);
+
+  g_hash_table_insert (xfm->from_atom_cache, (gpointer)atom, name2);
+
+  return name2;
+}
